# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from time import sleep
from datetime import datetime, timedelta
from Queue import Queue

from amiclient import AMIClient
from server import MainListener
from squirm import squirm, burials
from allios import readall, sendall
from config import read_config
from utils import tmp_debug

_1S = timedelta(0, 1)
_10S = timedelta(0, 10)


class Octopasty(object):

    def __init__(self, config_filename):
        self.config = read_config(config_filename)
        self.servers = self.config.get('amis')
        self.connect_tentatives = dict()
        self.amis = dict()
        self.clients = dict()
        self.in_queue = Queue()
        self.out_queue = Queue()
        self.flow = dict()
        self.listener = None

    # AMI side
    def connect_server(self, server):
        parameters = self.servers.get(server)
        client = AMIClient(self, server, parameters)
        client.start()

    def connect_servers(self):
        now = datetime.now()
        connected_servers = self.connected_servers
        for server in self.servers:
            if server not in connected_servers:
                last_try = self.connect_tentatives.get(server)
                if last_try is None or \
                   now - last_try > _1S:
                    self.connect_tentatives[server] = now
                    self.connect_server(server)

    def _get_connected_servers(self):
        return [ami.server for ami in \
                filter(lambda _ami: _ami.connected, self.amis.values())]
    connected_servers = property(_get_connected_servers)

    def _get_ami_sockets(self):
        return [ami.socket for ami in \
                filter(lambda _ami: _ami.connected, self.amis.values())]
    ami_sockets = property(_get_ami_sockets)

    # Client side
    def listen_clients(self):
        self.listener = MainListener(self)

    def _get_connected_clients(self):
        return [client.id for client in \
                filter(lambda _client: _client.connected,
                       self.clients.values())]
    connected_clients = property(_get_connected_clients)

    def _get_client_sockets(self):
        return [client.socket for client in \
                filter(lambda _client: _client.connected,
                       self.clients.values())]
    client_sockets = property(_get_client_sockets)

    def _get_peers(self):
        return list(self.clients.values()) + list(self.amis.values())
    peers = property(_get_peers)

    def find_peer_from_socket(self, s):
        for peer in self.peers:
            if peer.socket == s:
                return peer
        return None

    def get_peer(self, name):
        return self.amis.get(name) or self.clients.get(name) or None

    def idle(self):
        if len(self.connected_servers) == 0 and \
               len(self.connected_clients) == 0:
            sleep(1)

    def loop(self):
        looping = True
        self.listen_clients()
        while looping:
            try:
                self.connect_servers()
                self.idle()
                readall(self)
                squirm(self)
                burials(self)
                sendall(self)
            except KeyboardInterrupt:
                for ami in self.amis.values():
                    tmp_debug("NETWORK", "Disconecting from %s" % ami.uid)
                    ami.disconnect()
                    ami.join()
                for client in self.clients.values():
                    tmp_debug("NETWORK", "Disconecting %s" % client.uid)
                    client.disconnect()
                    client.join()
                looping = False
        self.listener.stop()
        self.listener.join()



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from utils import bigtime
from time import time
from copy import copy

from utils import Packet, tmp_debug, deprotect
from asterisk import Error
from internal import handle_action, KEEP_INTERNAL


def burials(self):
    apocalypse = bigtime(-10)
    for peer in self.peers:
        if peer and peer.locked and peer.logged and peer.locked < apocalypse:
            peer.locked = 0
            tmp_debug("SQUIRM", "Removing the very old lock of  %s" % \
                      peer.uid)


# The nervous central
# Thanks to Derek for the awesome function name !
def squirm(self):
    apocalypse = time() - 10
    if not self.in_queue.empty():
        incoming = list(self.in_queue.queue)
        self.in_queue.queue.clear()
        for packet in incoming:
            tmp_debug("SQUIRM", "Processing packet %s" % deprotect(packet))
            if packet.timestamp < apocalypse:
                # not puting in the queue and just continuing
                # makes the packet go nowhere
                # TODO: figure what to do here, i was thinking about something
                # if packet.locked:
                #     dest = self.flow.get("%s" % packet.locked)
                #     if dest != '__internal__':
                #         peer = self.get_peer(dest)
                tmp_debug("SQUIRM", "Removing very old packet: %s" % packet)
                continue
            if packet.emiter == '__internal__':
                peer = self.get_peer(packet.dest)
                if peer and peer.available:
                    self.out_queue.put(packet)
                else:
                    self.in_queue.put(packet)
            else:
                if packet.locked:
                    if packet.emiter in self.connected_servers:
                        tmp_debug("SQUIRM", "Emiter is server %s" % \
                                  packet.emiter)
                        emiter = self.get_peer(packet.emiter)
                        if emiter.locked and not emiter.keep_flow:
                            emiter.locked = 0
                            tmp_debug("SQUIRM", "Unlocking emiter %s" % \
                                      emiter.uid)
                        dest = self.flow.get("%s" % packet.locked)
                        if dest:
                            if dest == '__internal__':
                                handle_action(self, packet)
                            else:
                                peer = self.get_peer(dest)
                                if peer and peer.logged:
                                    if peer.locked:
                                        if peer.locked == packet.locked:
                                            packet.dest = dest
                                            self.out_queue.put(packet)
                                        else:
                                            tmp_debug("SQUIRM",
                                                      "Requeueing packet " \
                                                      "from server: %s" % \
                                                      deprotect(packet))
                                            self.in_queue.put(packet)
                                else:
                                    # just forgot it, and stop flow if any
                                    if emiter.keep_flow:
                                        emiter.keep_flow = 0
                                        emiter.locked = 0
                        else:
                            if emiter.keep_flow:
                                emiter.keep_flow = 0
                                emiter.locked = 0
                    if packet.emiter in self.connected_clients:
                        tmp_debug("SQUIRM", "Emiter is client %s" % \
                                  packet.emiter)
                        if packet.packet.name in KEEP_INTERNAL:
                            handle_action(self, packet)
                            continue
                        client = self.get_peer(packet.emiter)
                        if client.logged:
                            cid = client.id.split('_')[0]
                            cu = self.config['users'].get(cid)
                            if cu:
                                cs = cu.get('server')
                                if cs:
                                    peer = self.get_peer(cs)
                                    if peer and peer.logged:
                                        if peer.available:
                                            # server seems already available
                                            # but may be busy just after,
                                            # then we need to verify that
                                            # in outgoing packets too
                                            packet.dest = cs
                                            self.out_queue.put(packet)
                                            tmp_debug("SQUIRM",
                                                      "Packet from client " \
                                                      "pushed to queue %s" % \
                                                      deprotect(packet))
                                        else:
                                            # server seems busy
                                            self.in_queue.put(packet)
                                            tmp_debug("SQUIRM",
                                                      "Requeueing packet from "
                                                      " client: %s" % \
                                                      deprotect(packet))
                                    else:
                                        response = Error(
                                            parameters=dict(
                                                Message='Server unavailable'))
                                        p = dict(emiter=cs,
                                                 locked=packet.locked,
                                                 timestamp=time(),
                                                 packet=response,
                                                 dest=client.id)
                                        self.out_queue.put(Packet(p))
                        else:
                            packet.dest = '__internal__'
                            self.out_queue.put(packet)
                else:
                    # packet not locked
                    if packet.emiter in self.connected_servers:
                        # broadcast events
                        ami = self.get_peer(packet.emiter)
                        if ami.logged:
                            for client in self.connected_clients:
                                peer = self.get_peer(client)
                                if peer.binded_server == packet.emiter and \
                                   peer.wants_events:
                                    p = copy(packet)
                                    p.dest = client
                                    self.out_queue.put(p)
                        else:
                            # Do not remove that comment and that case
                            # it might be logged one day
                            # ami not logged, and no command before
                            # i wonder what can it be
                            pass



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import socket
from threading import Thread
from time import sleep, time
import errno
from asterisk import Action
from asterisk import STARTING_EVENTS_KEYWORDS

from utils import Packet, bigtime
from utils import deprotect, tmp_debug

LISTENING_TIMEOUT = 10


class ServerThread(Thread):

    def __init__(self, octopasty, channel, details):
        Thread.__init__(self)
        self.octopasty = octopasty
        self.socket = channel
        self.details = details
        self.action = None
        self.logged = False
        self.id = 'unknown_%d' % bigtime()
        self.locked = 0
        self.wants_events = False
        self.binded_server = None
        self.keep_flow = False
        self.buffer = ''
        self.lines = list()

    def disconnect(self):
        if self.socket:
            try:
                self.socket.shutdown(socket.SHUT_RDWR)
                self.socket.close()
            except:
                pass
        if self.id in self.octopasty.clients:
            self.octopasty.clients.pop(self.id)
        self.connected = False
        self.socket = None

    def run(self):
        self.connected = True
        self.socket.sendall("Asterisk Call Manager/1.1\r\n")
        self.octopasty.clients.update({self.id: self})

    def send(self, packet):
        released_lock = None
        if self.locked:
            if packet.locked == self.locked:
                tmp_debug("IO", "O => %s: %s" % (self.uid,
                                        deprotect(packet.packet)))
                try:
                    self.socket.sendall(str(packet.packet))
                except:
                    self.disconnect()
                    return None
                released_lock = self.locked
                if not self.keep_flow:
                    self.locked = 0
            else:
                # just discard, may be too old
                pass
        else:
            # on let events go
            if not packet.locked:
                tmp_debug("IO", "O => %s %s" % (self.uid,
                                        deprotect(packet.packet)))
                self.socket.sendall(str(packet.packet))
            else:
                # humm, why we get that ??
                pass
        return released_lock

    def handle_line(self):
        if self.socket:
            bytes = ''
            try:
                bytes = self.socket.recv(4096)
            except:
                pass
            if len(bytes) == 0:
                self.disconnect()
            self.buffer += bytes
        in_lines = self.buffer.split('\n')
        self.lines.extend([l + '\n' for l in in_lines[:-1]])
        self.buffer = in_lines[-1]
        for line in self.lines:
            tmp_debug("IO", "%s => O: %s" % (self.uid, deprotect(line)))
            line = line.strip()
            # if locked, we are waiting for a result
            if not self.locked:
                if line.lower().startswith('action:'):
                    self.action = \
                             Action(line[line.find(':') + 1:].strip())
                elif line == '':
                    if self.action:
                        self.locked = bigtime()
                        self.push(self.action)
                        self.action = None
                else:
                    if ':' in line:
                        k = line.split(':')[0]
                        v = line[line.find(':') + 1:].strip()
                    if self.action:
                        self.action.add_parameters({k: v})
        self.lines = []

    def push(self, packet):
        if packet.name.lower() in STARTING_EVENTS_KEYWORDS:
            self.keep_flow = True
        p = dict(emiter=self.id, locked=self.locked,
                 timestamp=time(), packet=packet)
        tmp_debug("PACKET", "%s (L:%s, KF:%s) >> %s" % (self.uid, self.locked,
                                                        self.keep_flow,
                                                        deprotect(str(p))))
        self.octopasty.in_queue.put(Packet(p))

    def _get_available(self):
        return self.connected and not self.locked
    available = property(_get_available)

    def _get_uid(self):
        return self.id
    uid = property(_get_uid)


class MainListener(Thread):

    def __init__(self, octopasty):
        Thread.__init__(self)
        self.octopasty = octopasty
        self.octopasty.listener = self
        self.start()

    def run(self):
        self.listening = True
        while self.listening:
            try:
                self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.s.bind((self.octopasty.config.get('bind_address'),
                             int(self.octopasty.config.get('bind_port'))))
                tmp_debug("NETWORK", "Listening on %s %s" % \
                          (self.octopasty.config.get('bind_address'),
                           self.octopasty.config.get('bind_port')))
                self.s.listen(5)
                while self.listening:
                    channel, details = self.s.accept()
                    st = ServerThread(self.octopasty, channel, details)
                    st.start()
            except socket.error, e:
                if e.errno == errno.EADDRINUSE:
                    tmp_debug(
                        "NETWORK",
                        "Address already used trying to bind on %s %s" % \
                        (self.octopasty.config.get('bind_address'),
                         self.octopasty.config.get('bind_port')))
                    for timeout in range(0, LISTENING_TIMEOUT):
                        if self.listening == False:
                            break
                        sleep(1)
                elif e.errno == errno.EINVAL:
                    tmp_debug("NETWORK", "Closing server, time to sleep")
                    self.listening = False
                else:
                    tmp_debug("NETWORK", "Unknown socket error: %s" % e)
                    self.listening = False

    def stop(self):
        try:
            self.s.shutdown(socket.SHUT_RDWR)
        except:
            # happend with there is already a listener
            self.listening = False



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from hashlib import sha1
from time import time

from utils import Packet, bigtime, tmp_debug
from asterisk import Success, Error, Goodbye

KEEP_INTERNAL = ['logoff']


def handle_action(self, packet):

    action = packet.packet
    if action.name.lower() == 'error':
        # needs to handle errors, may be a timeout before next try
        login_failed_on_ami(self, packet.emiter)

    if action.name.lower() == 'success':
        logged_on_ami(self, packet.emiter)

    if action.name.lower() == 'login':
        login = dict()
        for k in ['Username', 'Secret', 'Events']:
            v = action.parameters.get(k) or \
                action.parameters.get(k.lower()) or \
                action.parameters.get(k.upper())
            login[k.lower()] = v
        auth_user(self, packet.emiter, packet.locked, login.get('username'),
                  login.get('secret'),
                  login.get('events') and \
                  login.get('events').lower() == 'on' or False)

    if action.name.lower() == 'logoff':
        logoff_user(self, packet)


def auth_user(self, emiter, locked, username, secret, wants_events):
    if username in self.config.get('users'):
        hashed = self.config.get('users').get(username).get('password')
        client = self.clients.get(emiter)
        if sha1(secret).hexdigest() == hashed:
            old_id = client.id
            client.id = '%s_%d' % (username, bigtime())
            self.clients.pop(old_id)
            self.clients.update({client.id: client})
            client.logged = True
            server = self.config.get('users').get(username).get('server')
            client.binded_server = server
            client.wants_events = wants_events
            response = Success(parameters=dict(
                Message='Authentication accepted'))
            p = dict(emiter='__internal__',
                     locked=locked,
                     timestamp=time(),
                     packet=response,
                     dest=client.id)
            tmp_debug("AUTH", "'%s' logged successfully" % username)
            self.out_queue.put(Packet(p))
        else:
            response = Error(parameters=dict(Message='Authentication failed'))
            p = dict(emiter='__internal__',
                     locked=locked,
                     timestamp=time(),
                     packet=response,
                     dest=client.id)
            client.send(Packet(p))
            tmp_debug("AUTH", "'%s' failed to login" % username)
            client.disconnect()


def logoff_user(self, packet):
    client = self.clients.get(packet.emiter)
    response = Goodbye(parameters=dict(Message="Don't panic."))
    p = dict(emiter='__internal__',
             locked=packet.locked,
             timestamp=time(),
             packet=response,
             dest=client.id)
    client.send(Packet(p))
    tmp_debug("AUTH", "'%s' logout" % packet.emiter[:packet.emiter.find('_')])
    client.disconnect()


def login_failed_on_ami(self, _ami):
    tmp_debug("AUTH", "Login failed on '%s'" % _ami)


def logged_on_ami(self, _ami):
    tmp_debug("AUTH", "Logged on '%s'" % _ami)
    ami = self.amis.get(_ami)
    ami.logged = True



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import socket
from threading import Thread
from time import time
from utils import Packet, bigtime
from utils import deprotect, tmp_debug

from asterisk import Login
from asterisk import Event, Response
from asterisk import STARTING_EVENTS_KEYWORDS
from asterisk import STOPPING_EVENTS_KEYWORDS


class AMIClient(Thread):
    def __init__(self, octopasty, server, parameters):
        Thread.__init__(self)
        self.octopasty = octopasty
        self.server = server
        self.socket = None
        self.host = parameters.get('host')
        self.port = parameters.get('port')
        self.user = parameters.get('user')
        self.password = parameters.get('password')
        self.connected = False
        self.response = None
        self.event = None
        self.logged = False
        self.locked = False
        self.keep_flow = False
        self.buffer = ''
        self.lines = []

    def disconnect(self):
        if self.socket:
            try:
                self.socket.shutdown(socket.SHUT_RDWR)
                self.socket.close()
            except:
                pass
        if self.server in self.octopasty.amis:
            self.octopasty.amis.pop(self.server)
        self.connected = False
        self.socket = None

    def send(self, packet):
        if not self.locked:
            tmp_debug("IO", "O => %s: L: %s D: %s" % (
                self.uid,
                packet.locked,
                deprotect(packet.packet)))
            self.socket.sendall(str(packet.packet))
            if packet.packet.name.lower() in STARTING_EVENTS_KEYWORDS:
                self.keep_flow = True
                tmp_debug("PACKET", "%s Starting flow (L:%s)" % (self.uid,
                                                                 self.locked))
            self.locked = packet.locked
            return self.locked
        else:
            return None

    def login(self):
        self.push(Login(self.user, self.password),
                  emiter='__internal__', dest=self.server,
                  locked=bigtime())

    def run(self):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((self.host, int(self.port)))
            self.socket = s
            self.octopasty.amis.update({self.server: self})
            self.connected = True
        except socket.error:
            pass

    def handle_line(self):
        if self.socket:
            bytes = ''
            try:
                bytes = self.socket.recv(4096)
            except:
                pass
            if len(bytes) == 0:
                self.disconnect()
            self.buffer += bytes
        in_lines = self.buffer.split('\n')
        self.lines.extend([l + '\n' for l in in_lines[:-1]])
        self.buffer = in_lines[-1]
        for line in self.lines:
            tmp_debug("IO", "%s => O: L: %s D: %s" % (
                self.uid, self.locked, deprotect(line)))
            line = line.strip()
            if line.startswith('Asterisk Call Manager'):
                self.login()
                self.lines = []
                return
            if self.locked and line.lower().startswith('response:'):
                self.response = \
                          Response(line[line.find(':') + 1:].strip())
                self.event = None
            elif line.lower().startswith('event:'):
                name = line[line.find(':') + 1:].strip()
                self.event = Event(name)
                self.response = None
            elif line == '':
                if self.response:
                    self.push(self.response)
                    self.response = None
                elif self.event:
                    if self.keep_flow:
                        self.push(self.event)
                    else:
                        self.push(self.event, locked=0)
                    self.event = None
            else:
                if ':' in line:
                    sc = line.find(':')
                    k = line[:sc].strip()
                    v = line[sc + 1:].strip()
                else:
                    k = line
                    v = None
                if self.response:
                    self.response.add_parameters({k: v})
                if self.event:
                    self.event.add_parameters({k: v})
        self.lines = []

    def push(self, packet, emiter=None, dest=None, locked=-1):
        p = dict(emiter=emiter or self.server,
                 locked=locked == -1 and self.locked or locked,
                 timestamp=time(),
                 packet=packet)
        if dest:
            p['dest'] = dest
        tmp_debug("PACKET", "%s (L:%s, KF:%s) >> %s" % (self.uid, self.locked,
                                                        self.keep_flow,
                                                        deprotect(p)))
        self.octopasty.in_queue.put(Packet(p))
        if packet.name.lower() in STOPPING_EVENTS_KEYWORDS:
            self.keep_flow = False
        tmp_debug("PACKET", "%s Stopping flow (L:%s)" % (self.uid,
                                                         self.locked))

    def _get_available(self):
        return self.connected and not self.locked
    available = property(_get_available)

    def _get_uid(self):
        return self.server
    uid = property(_get_uid)



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ConfigParser import ConfigParser


def get_amis(config):
    ret = dict()
    ret['amis'] = dict()
    assert config.has_section("amis"), \
           "No 'amis' section in the config file"
    amis = config.items('amis')
    assert len(amis) > 0, "No AMIs configured"
    for key, parameters in amis:
        left, right = parameters.split('@')
        user, password = left.split(':')
        host, port = right.split(':')
        ret['amis'][key] = dict(user=user, password=password,
                                host=host, port=port)
    return ret


def get_server(config):
    ret = dict()
    assert config.has_section("server"), \
           "No 'server' section in the config file"
    ret['bind_address'] = config.get('server', 'address')
    ret['bind_port'] = config.get('server', 'port')
    return ret


def get_users(config):
    ret = dict()
    ret['users'] = dict()
    assert config.has_section("users"), \
           "No 'users' section in the config file"
    users = config.items('users')
    assert len(users) > 0, "No AMIs configured"
    for key, parameters in users:
        parameters = parameters.replace(' ', '')
        password, server = parameters.split('@')
        user = key
        ret['users'][user] = dict(password=password,
                                  server=server)
    return ret


def read_config(filename):
    config = dict()
    parser = ConfigParser()
    parser.read(filename)
    config.update(get_amis(parser))
    config.update(get_server(parser))
    config.update(get_users(parser))
    return config



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from select import select
from internal import handle_action
from asterisk import STOPPING_EVENTS_KEYWORDS
from utils import tmp_debug, deprotect


def readall(self):
    reading = self.ami_sockets + self.client_sockets
    if len(reading) > 0:
        # it's a timeout, that means that 0 != None
        timeout = self.out_queue.empty() and 0.1 or 0
        to_read, _, _ = select(reading, [], [], timeout)
        for s in to_read:
            peer = self.find_peer_from_socket(s)
            peer.handle_line()


def sendall(self):
    if not self.out_queue.empty():
        outgoing = list(self.out_queue.queue)
        self.out_queue.queue.clear()
        for packet in outgoing:
            tmp_debug("PACKET", "%s << %s" % (packet.dest, deprotect(packet)))
            if packet.dest == '__internal__':
                handle_action(self, packet)
            else:
                dest = self.get_peer(packet.dest)
                if dest:
                    # here is another place to check
                    # and requeue if necessary
                    if packet.locked != dest.locked and \
                       not dest.available:
                        self.out_queue.put(packet)
                        tmp_debug("SQUIRM", "Requeueing on output: %s" % \
                                  deprotect(packet))
                        continue
                    if not packet.locked and not dest.available:
                        self.out_queue.put(packet)
                        tmp_debug("SQUIRM", "Requeueing on output: %s" % \
                                  deprotect(packet))
                        continue
                    sent = dest.send(packet)
                    if sent:
                        if str(sent) in self.flow:
                            keep_flow = dest.keep_flow
                            if packet.packet.name.lower() in \
                                   STOPPING_EVENTS_KEYWORDS:
                                keep_flow = False
                                dest.keep_flow = False
                                dest.locked = 0
                            # then it was an answer
                            if not keep_flow:
                                self.flow.pop("%s" % sent)
                        else:
                            # then it was a query
                            self.flow["%s" % sent] = packet.emiter
                    else:
                        # Then it was an event
                        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup
from octopasty import __version__

packages = ['octopasty']

scripts = """ [console_scripts]
              octopasty = octopasty.run:main
              """

setup(name='octopasty',
      version=__version__,
      description="Octopasty Asterisk AMI proxy",
      author="Jean Schurger",
      author_email='jean@schurger.org',
      url='http://labs.ubity.com/projects/octopasty',
      packages=packages,
      license='GPLv3',
      entry_points=scripts,
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from optparse import OptionParser

from octopasty import __version__
from octopasty.main import Octopasty


def main():
    parser = OptionParser(usage="%prog CONFIG_FILE",
                          description=__doc__,
                          version=__version__)
    args = parser.parse_args()[1]
    if len(args) < 1:
        parser.error("I need config file a config file as argument")
    o = Octopasty(args[0])
    o.loop()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# must be lowercased
STARTING_EVENTS_KEYWORDS = ['queuestatus']
STOPPING_EVENTS_KEYWORDS = ['queuestatuscomplete']


class MetaPacket(object):
    _name_ = 'MetaPacket'
    _direction_ = 'Action'

    def __init__(self, name=None, parameters=None):
        self.parameters = dict()
        self.parameters_order = list()
        self.name = name or self._name_
        if self.name and parameters:
            self.add_parameters(parameters)

    def add_parameters(self, parameters):
        self.parameters.update(parameters)
        self.parameters_order.extend(parameters.keys())

    def __repr__(self):
        out = "%s: %s\r\n" % (self._direction_, self.name)
        for p in self.parameters_order:
            if self.parameters[p] == None:
                out += "%s\n" % p
                pass
            else:
                out += "%s: %s\r\n" % (p, self.parameters.get(p))
        out += "\r\n"
        return out


# main classes
class Event(MetaPacket):
    _name_ = 'Event'
    _direction_ = 'Event'


class Response(MetaPacket):
    _name_ = 'Response'
    _direction_ = 'Response'


class Action(MetaPacket):
    _name_ = 'Action'


# real things
class Login(Action):
    _name_ = 'Login'

    def __init__(self, login, password):
        Action.__init__(self, parameters=dict(Username=login,
                                               Secret=password,
                                               Events='on'))


class Success(Response):
    _name_ = 'Success'


class Error(Response):
    _name_ = 'Error'


class Goodbye(Response):
    _name_ = 'Goodbye'



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from time import time
from datetime import datetime


class NiceDict(object):
    def __init__(self, entries):
        self.__dict__.update(entries)


class Packet(NiceDict):

    def __repr__(self):
        ret = ""
        ret += "E: %s, " % self.emiter
        ret += "T: %s, " % self.timestamp
        ret += "L: %s, " % self.locked
        if hasattr(self, 'dest'):
            ret += "D: %s, " % self.dest
        ret += "P: %s" % \
               str(self.packet).replace('\n', '\\n').replace('\r', '\\r')
        ret += "\n"
        return ret


def bigtime(delta=0):
    return int((time() + delta) * 10000000)


def deprotect(s):
    return str(s).replace('\r', '\\r').replace('\n', '\\n')


def tmp_debug(t, s):
    debug = os.environ.get('OCTOPASTY_DEBUG')
    debug = debug and debug.split(',') or []
    if t in debug:
        print "@%s [% 8s] %s" % (datetime.now(), t, s)
